import logging
import logging.handlers
import os
import sys
from getpass import getuser
from pickle import dump, load
from queue import Queue
from re import match
from tkinter import BOTH, Menu, PhotoImage, StringVar, Tk, messagebox

import eaiConstants
from logger import setup_logging
from ui.login_frame import LoginFrame
from ui.main_frame import MainFrame
from ui.utils import find_data_file
from version import eggPlantAIAgentBuild, versionOrBuild


class AgentGui(object):
    """ A GUI for the Eggplant AI Agent written using TkInter
    """

    def __init__(self, queue):
        """ GUI Constructor
        """
        self.persistentData = {}

        self.root = Tk()
        self.root.title(
            "Eggplant AI Agent ({} - {})".format(versionOrBuild(), eggPlantAIAgentBuild)
        )

        try:
            if sys.platform.startswith("linux"):
                img = PhotoImage(file=find_data_file("resources/AI-circle-64.gif"))
                self.root.tk.call("wm", "iconphoto", self.root._w, img)
            else:
                self.root.wm_iconbitmap(
                    find_data_file("resources/ai_circle_x64-x16.ico")
                )
        except Exception:
            print("Could not load icon.")

        self.root.columnconfigure(0, weight=1)
        self.root.rowconfigure(0, weight=1)
        self.root.minsize(800, 768)

        # Custom about menu for Mac OS.
        if sys.platform == "darwin":
            self.menubar = Menu(self.root)
            self.app_menu = Menu(self.menubar, name="apple")
            self.menubar.add_cascade(menu=self.app_menu)
            self.app_menu.add_command(
                label="About Eggplant AI Agent", command=self.do_about_dialog
            )
            self.app_menu.add_separator()
            self.root.config(menu=self.menubar)

        self.persistentDataDefaults = {
            "connectionName": getuser(),
            "driveExec": eaiConstants.DEFAULT_EGGDRIVE_EXECUTABLE,
            "port": eaiConstants.DEFAULT_EGGDRIVE_PORT,
            "debug": "",
            "epFGUI": "1",
            "cleanEpf": "",
            "ssl": "1",
            "serverName": eaiConstants.DEFAULT_SERVER,
            "designMode": "1",
            "suiteRootFolder": "",
            "token": "",
        }
        self.persistentDataFile = os.path.join(
            str(os.path.expanduser("~")), ".eggplantAIagent.ini"
        )
        self.loadPersistentData()

        self.mainframe = MainFrame(self.persistentData, queue, self.root, self.logout)

        self.login_frame = LoginFrame(
            self.persistentData["serverName"],
            self.persistentData["ssl"],
            self.persistentData["token"],
            self.login,
            self.root,
        )

        self.displayLoginFrame()

    def __del__(self):
        self.finish()

    def do_about_dialog(self):
        messagebox.showinfo(
            message="Eggplant AI Agent\n\nVersion: {}".format(versionOrBuild())
        )

    def finish(self):
        """ GUI destructor
        """
        self.savePersistentData()
        self.root.quit()

    def loadPersistentData(self):
        """ Load persistent data from a file using pickle
        """
        self.persistentData = {}
        # default values for the saved data
        saveData = ""
        print("Loading %s" % (self.persistentDataFile,))
        if os.path.isfile(self.persistentDataFile):
            try:
                with open(self.persistentDataFile, "rb") as file:
                    saveData = load(file)
            except Exception:
                os.remove(self.persistentDataFile)

        # go through each of the required parameters and either set it
        # from loaded data or defaults
        for key, val in self.persistentDataDefaults.items():
            if (type(saveData) is dict) and (key in saveData):
                self.persistentData[key] = StringVar(value=saveData[key])
            else:
                self.persistentData[key] = StringVar(value=val)

    def savePersistentData(self):
        """
        Save persistent data to a file using pickle
        """
        self.checkPersistentdata()
        saveData = {}
        for key in self.persistentData:
            saveData[key] = self.persistentData[key].get()

        with open(self.persistentDataFile, "wb") as file:
            dump(saveData, file)

    def startGui(self):
        """ Start the GUI
        """

        # connect a process to the window close event
        self.root.protocol("WM_DELETE_WINDOW", self.finish)

        # start the GUI running
        self.root.focus()
        self.root.mainloop()

    def displayLoginFrame(self):
        self.mainframe.pack_forget()
        self.login_frame.pack(fill=BOTH, expand=1)

    def displayMainFrame(self):
        self.login_frame.pack_forget()
        self.mainframe.pack(fill=BOTH, expand=1)

    def login(self):
        self.mainframe.on_logged_in()
        self.displayMainFrame()

    def logout(self):
        self.displayLoginFrame()

    def checkPersistentdata(self):
        """ Check the validity of various options
        """
        # remove unwanted http:\\ or https:\\ stuff from around the server name
        validHostRegex = r"(?:http(?:s)?:\/\/)?([^\/\r\n]+)(?:\/[^\r\n]*)?"
        m = match(validHostRegex, self.persistentData["serverName"].get())
        if m and (len(m.group(1)) > 0):
            self.persistentData["serverName"].set(m.group(1))
        self.persistentData["connectionName"].set(
            self.persistentData["connectionName"].get().strip()
        )


if __name__ == "__main__":

    log_queue = Queue()

    setup_logging(
        logging.DEBUG, "%(asctime)s [%(levelname)-5.5s] %(message).10000s", log_queue
    )

    gui = AgentGui(log_queue)
    gui.startGui()
