import asyncio
import threading
from enum import Enum


class DownloadStatus(Enum):
    NOT_STARTED = 0
    IN_PROGRESS = 1
    COMPLETED = 2
    FAILED = 3
    CANCELLED = 4


class DownloadResult(object):
    def __init__(self, status=DownloadStatus.NOT_STARTED):
        self.status = status
        self.exception = None
        self.version = None
        self.lock = threading.Lock()

    def get_status(self):
        with self.lock:
            return self.status

    def update_status(self, status, exception=None, version=None):
        with self.lock:
            self.status = status
            self.exception = exception
            self.version = version

    def to_dict(self):
        with self.lock:
            return {
                "status": self.status.value,
                "error_message": str(self.exception) if self.exception else None,
                "version": self.version.dump() if self.version else None,
            }


class AssetController(object):
    """
    Downloads assets from the VAM on a background thread
    """

    def __init__(self, asset_manager):
        self.asset_manager = asset_manager
        self.download_complete_callback = None
        self.thread = None
        self.loop = None
        self.is_loop_running = threading.Event()
        self.future = None
        self.result = DownloadResult()

    def start(self):
        """Start the background thread"""
        self.thread = threading.Thread(target=self.run)
        self.thread.start()

        # Wait for the event loop to be started before returning
        self.is_loop_running.wait()

    def stop(self):
        """Stop the background thread"""
        # Tell the event loop to stop
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join()

    def run(self):
        """Run an asyncio event loop"""

        # Create a new event loop for this thread
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        # As soon as the loop has started, set the is_loop_running event
        self.loop.call_soon(self.is_loop_running.set)

        # Loop until told to stop
        self.loop.run_forever()
        self.loop.close()

    def download_asset(self, assets_dir, version_id):
        """Start downloading an asset from the VAM"""

        if self.result.get_status() == DownloadStatus.IN_PROGRESS:
            return

        self.result.update_status(DownloadStatus.IN_PROGRESS)

        # Start downloading the asset on the event loop thread
        self.future = asyncio.run_coroutine_threadsafe(
            self.asset_manager.download_asset_async(assets_dir, version_id=version_id),
            self.loop,
        )

        self.future.add_done_callback(self.on_download_complete)

    def on_download_complete(self, future):
        try:
            version = future.result()
            self.result.update_status(DownloadStatus.COMPLETED, version=version)
        except asyncio.CancelledError:
            self.result.update_status(DownloadStatus.CANCELLED)
        except Exception as e:
            self.result.update_status(DownloadStatus.FAILED, exception=e)

        if self.download_complete_callback is not None:
            self.download_complete_callback(self.result)

    def cancel_download(self):
        if self.result.get_status() != DownloadStatus.IN_PROGRESS:
            self.result.update_status(DownloadStatus.CANCELLED)
        elif self.future is not None:
            self.future.cancel()

    def get_result(self):
        return self.result.to_dict()
