import logging
import os
import sys
import psutil
import signal
import platform
import uuid
import pathlib

from threading import Thread

import subprocess

from xmlrpc.client import ServerProxy, dumps, Fault

import re

cur_version = sys.version_info


class EggDriveStarter(object):
    """
    Provides the interface to eggDrive starting it on instantiation and
    passing commands and returning results as necessary
    """

    def __init__(self, driveExec, port, epfgui, cleanupepf, debugEpf, startEpf):
        """Start up eggDrive and open the xmlrpc interface
        """
        self.eggDriveCommand = []
        self.eggDriveCommand.append(str(driveExec))
        # add the gui command if required
        self.eggDriveCommand.append("-driveport")
        self.eggDriveCommand.append(str(port))
        if epfgui:
            self.eggDriveCommand.append("-runGUI")
            self.eggDriveCommand.append("Yes")

        self.eggDriveCommand.append("-RedirectOutputToFile")
        self.eggDriveCommand.append("Yes")

        # extra drive interface logging information from ePF
        if debugEpf:
            self.eggDriveCommand.append("-drivelogging")
            self.eggDriveCommand.append("1")

        self.defaultTimeout = 10

        # Check if there are any ePF processes still running and remove them
        if cleanupepf:
            self.removeZombieEpfProcesses()

        if startEpf:
            self.thread = Thread(target=self.startEpf)
            self.thread.start()

        self.proxy = ServerProxy("http://127.0.0.1:" + str(port) + "/", verbose=False)

    def display_epf_command(self):
        msg = 'Starting Eggplant Functional with command: "'
        for c in self.eggDriveCommand:
            msg = "{0} {1}".format(msg, c)
        msg = '{0}"'.format(msg)

        logging.info(msg)

    def startEpf(self):
        try:
            self.display_epf_command()

            # Running pyinstaller with the "-windowed" option on windows
            # can cause some problems.
            # Use the recipe here:
            # https://github.com/pyinstaller/pyinstaller/wiki/Recipe-subprocess
            # to avoid them
            if hasattr(subprocess, "STARTUPINFO"):
                # Only on windows
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                env = os.environ
            else:
                startupinfo = None
                env = None

            self.eggDriveProcess = subprocess.Popen(
                self.eggDriveCommand,
                shell=False,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                startupinfo=startupinfo,
                env=env,
            )
        except Exception:
            logging.exception("Unable to start EPF")
            return

        # Poll process for new output until finished
        while self.eggDriveProcess.poll() is not None:
            self.eggDriveProcess.stdout.flush()
            nextline = self.eggDriveProcess.stdout.readline()

            if nextline.strip() != "":
                logging.info("Functional: {}".format(nextline))

        output = self.eggDriveProcess.communicate()[0]
        exitCode = self.eggDriveProcess.returncode

        if exitCode == 0:
            return output
        else:
            pass

    def xml_rpc_call(self, method_name, *args):
        """
        Make an XMLRPC call to EPF, logging the request and response
        """
        logging.debug(
            "Sending %s command to EPF:\n%s", method_name, dumps(args, method_name)
        )

        try:
            result = getattr(self.proxy, method_name)(*args)
            logging.debug(
                "Received from EPF:\n%s", dumps((result,), methodresponse=True)
            )

            return result
        except Fault as e:
            logging.debug("EPF error:\n%s", dumps(e, methodresponse=True))
            raise

    def end_session(self, arg):
        return self.xml_rpc_call("EndSession", arg)

    def start_session(self, arg):
        return self.xml_rpc_call("StartSession", arg)

    def execute(self, arg):
        return self.xml_rpc_call("execute", arg)

    def execute_with_timeout(self, timeout, arg):
        return self.xml_rpc_call("ExecuteWithTimeout", timeout, arg)

    def update(self, timeout):
        return self.xml_rpc_call("Update", timeout)

    def capture_image(self, arg):
        return self.xml_rpc_call("CaptureImage", arg)

    def get_screen(self, arg):
        return self.xml_rpc_call("GetScreen", arg)

    def quit(self, arg):
        return self.xml_rpc_call("Quit", arg)

    def forceCloseSession(self, arg):
        # ensure the session is closed before sending the commands
        try:
            if "path" in arg:
                self.end_session(arg["path"])
            else:
                self.end_session("")
        except Exception:
            logging.debug("Failed to close session for {}".format(arg.get("path")))

    def send(self, cmd, arg):
        """ Send commands to eggDrive via XMLRPC
        """
        logging.info("Executing on EPF: {0}({1})".format(cmd, arg))
        result = ""
        resultDict = {}
        try:
            if cmd == "execute":
                resultDict = self.execute(arg)
                result = self.stringFromResponse(
                    resultDict
                )  # formulate a string result
            elif cmd == "script":

                self.forceCloseSession(arg)

                if "path" in arg:
                    self.start_session(arg["path"])
                else:
                    self.start_session("")

                resultDict = self.execute(arg["script"])
                result = self.stringFromResponse(
                    resultDict
                )  # formulate a string result

                if "path" in arg:
                    self.end_session(arg["path"])
                else:
                    self.end_session("")

            elif cmd == "ExecuteWithTimeout":
                resultDict = self.execute_with_timeout(self.defaultTimeout, arg)
                result = self.stringFromResponse(
                    resultDict
                )  # formulate a string result
            elif cmd == "Update":
                resultDict = self.update(self.defaultTimeout)
                result = self.stringFromResponse(
                    resultDict
                )  # formulate a string result
            elif cmd == "CaptureImage":
                logging.debug("Capture Image for path={0}".format(arg["path"]))

                self.forceCloseSession(arg)

                self.start_session(arg["path"])

                hosts = self.execute(
                    "put the AllConnectionInfo into HostList\nreturn HostList"
                )
                connected = list(
                    filter(lambda x: x["Status"] == "Connected", hosts["ReturnValue"])
                )

                if len(connected) == 1:
                    self.execute('connect "{0}"'.format(connected[0]["Name"]))
                    if "imageName" not in arg:
                        arg["imageName"] = "{}.png".format(uuid.uuid4())
                    result = self.capture_image(arg)
                    self.end_session(arg["path"])
                    return (
                        result,
                        {
                            "Output": "Image captured {0}.".format(arg["imageName"]),
                            "Result": result,
                        },
                    )
                self.end_session(arg["path"])

            elif cmd == "GetScreen":
                logging.debug("GetScreen path={0}".format(arg["path"]))

                self.forceCloseSession(arg)

                self.start_session(arg["path"])

                hosts = self.execute(
                    "put the AllConnectionInfo into HostList\nreturn HostList"
                )
                connected = list(
                    filter(lambda x: x["Status"] == "Connected", hosts["ReturnValue"])
                )

                if len(connected) == 1:
                    logging.debug("Hosts={0}".format(connected[0]["Name"]))
                    self.execute('connect "{0}"'.format(connected[0]["Name"]))
                    result = self.get_screen(arg)
                    self.end_session(arg["path"])

                    try:
                        data = {
                            "sutName": str(connected[0]["Name"]),
                            "data": result[0]["data"].data,
                            "width": int(result[0]["rectangle"][2]),
                            "height": int(result[0]["rectangle"][3]),
                        }
                    except Exception as e:
                        logging.error(
                            "Error receiving data: {}, {}".format(result, e.msg)
                        )
                        return (
                            result,
                            {
                                "Output": "Error capturing image",
                                "Result": {
                                    "sutName": str(connected[0]["Name"]),
                                    "data": "",
                                    "width": 0,
                                    "height": 0,
                                },
                            },
                        )

                    return (
                        result,
                        {
                            "Output": "SUT {0} screen captured.".format(
                                connected[0]["Name"]
                            ),
                            "Result": data,
                        },
                    )
                self.end_session(arg["path"])

            elif cmd == "StartSession":
                path = pathlib.Path(arg.replace("\\", "/"))

                if not path.exists():
                    raise Exception(
                        "Unable to StartSession: path {} does not exist".format(path)
                    )

                result = self.start_session(str(path))
            elif cmd == "EndSession":
                result = self.end_session("")
            else:
                logging.error("Command {0} not recognised".format(cmd))
        except Exception as e:
            logging.error("Call to EPF failed: %s", str(e))
            res = "Status=Failure, Errors=1.0, Successes=0.0, faultString={0}".format(
                str(e)
            )
            resultDict = {
                "Output": re.sub(r"[<>]", "", str(e)),
                "Result": {
                    "Status": "Failure",
                    "Errors": "1.0",
                    "Successes": "0.0",
                    "faultString": re.sub(r"[<>]", "", str(e)),
                },
            }
            res = re.sub(r"[<>]", "", res)
            return res, resultDict

        if "Output" in resultDict:
            logging.info(resultDict["Output"])
        return result, resultDict

    def getEpfZombies(self):
        """ get a list of all zombie processes rnning on this machine
        """
        zombies = []

        for proc in psutil.process_iter():
            try:
                pinfo = proc.as_dict(attrs=["pid", "name"])
            except psutil.NoSuchProcess:
                pass
            else:
                # TODO need to add MAC names to this list
                if pinfo["name"] in [
                    "runscript.exe",
                    "gdnc.exe",
                    "gpbs.exe",
                    "adb.exe",
                ]:
                    pinfo["process"] = proc
                    zombies.append(pinfo)

        return zombies

    def removeZombieEpfProcesses(self):
        """
        If the are zombie processes let over from a previous ePF run then we'll clean
        them here
        """
        zombies = self.getEpfZombies()

        if len(zombies) > 0:
            logging.info(
                "The following processes may be left over from a previous ePF run and "
                "will be cleaned:"
            )
            logging.info("    {0}".format([d["name"] for d in zombies]))
            for z in zombies:
                z["process"].kill()

    def filterEggplantFunctionalOutput(self, text):
        """ remove any unwanted text from the ePF output
        """
        lines = text.split("\n")
        lines = [
            re.sub(r"[<>]", "", l) for l in lines if not re.match(r"(^>>>)|(^<<<)", l)
        ]
        return "\n".join(lines)

    def stringFromResponse(self, response):
        """ We will formulate a string based response from the dictionary result
        from the xmlrpc executiom
        """
        result = ""
        for key, value in response.items():
            if key == "Output":
                value = self.filterEggplantFunctionalOutput(value)

            if key == "Result" and isinstance(value, dict):
                result += key + "="
                for keyi, valuei in value.items():
                    result += keyi + "=" + str(valuei) + ", "
            else:
                result += key + "=" + str(value) + ", "

        return result

    def killEggDrive(self):
        try:
            processId = self.eggDriveProcess.pid
            self.killChildProcesses(processId)
        except Exception:
            logging.exception("Failure during ending of Eggplant Functional processes")

    def killChildProcesses(self, parentPid, sig=signal.SIGTERM):
        logging.info("Killing Children of process {}".format(parentPid))

        try:
            parent = psutil.Process(parentPid)
        except psutil.NoSuchProcess:
            logging.info("No such process {}".format(parentPid))
            return

        children = parent.children(recursive=True)
        for process in children:
            logging.debug("Killing Child process {}".format(process.pid))
            if platform.system() != "Windows":  # kill -9 on MAC
                os.system("kill -9 {0}".format(process.pid))

            try:
                process.send_signal(sig)

                logging.debug("Killed Child process {}".format(process.pid))
            except Exception:
                logging.debug("Failed to kill process {}".format(process.pid))

    def close(self):
        """Shutdown eggDrive and the XMLRPC server
        """
        try:
            logging.info("Quitting Eggplant Functional")
            resultDict = self.quit("")
            if resultDict != "Eggplant Drive Terminated":
                try:
                    self.killEggDrive()
                except Exception as e:
                    logging.warning("Unable to kill all child processes: {}".format(e))
        except ConnectionRefusedError:
            # EPF probably not running in the first place
            pass
        except Exception:
            logging.exception("Error quitting Eggplant Functional")
