import logging
import os
import base64

from collections.abc import Iterable
from re import finditer, sub
from datetime import datetime


class ImageHandling(object):
    """ handles the image processing requirements of the agent
    """

    def __init__(self):
        """Constructor
        """
        self.sessionDirectory = ""  # the directory where the session is
        self.uniqueSessionName = ""  # A unique name for teh session
        self.uniqId = 0

    def findLatestInSessionDirectory(self, fileName):
        """ Search for the latest file matching fileName in the session directory
        """
        latest_file = {"name": "", "ts": 0}
        for root, _dirs, files in os.walk(
            os.path.join(self.sessionDirectory, "Results")
        ):
            for f in files:
                if f == fileName:
                    p = os.path.join(root, f)
                    ts = os.path.getmtime(p)
                    if ts > latest_file["ts"]:
                        latest_file = {"name": p, "ts": ts}

        return latest_file["name"]

    def imageFileProcess(self, resultDict):
        """ Process the image filename in the output.
        If we see an image filename in the output from ePF then we need to upload it to
        the server.
        To do this we need to convert it to base64 and include it in the message that
        we send back. We also need to modify the name to reflect a server based name
        rather than the local file system  name. We use two strategies to look for the
        correct location for the file. In the first instance we use the logfile
        declaration in teh output which tells us where the loffile is currently being
        written. If this is not available then we search the session location for the
        latest file of the right name and this is the file that we will use.
        """
        images = {}  # a store for the images
        logDirectory = ""  # the directory where log files are currently being put

        # Get the current log directory
        if (
            "Result" in resultDict
            and isinstance(resultDict["Result"], Iterable)
            and "LogFile" in resultDict["Result"]
        ):
            try:
                logging.debug(
                    'imageFileProcess(resultDict["Result"]["LogFile"]={0})'.format(
                        str(resultDict["Result"]["LogFile"])
                    )
                )
                logDirectory, tail = os.path.split(resultDict["Result"]["LogFile"])
                logging.info("Using log directory {0}, {0}".format(logDirectory, tail))
            except Exception as e:
                logging.info("Error updating log path {0}".format(str(e)))
        else:
            logging.info('Using existing log path "{0}"'.format(logDirectory))
            logging.debug("imageFileProcess(resultDict={0})".format(str(resultDict)))

        # get a list of all png files
        if "Output" in resultDict:
            try:
                # Find file references in the log
                regex = (
                    r"(?:Log|FAILURE|capturescreen|moveto|tracescreen)\s+(\w.*\.png)"
                )
                # regex = r'\s(\S*\.png)\s'
                for f in finditer(regex, resultDict["Output"]):
                    fullPathAndName = f.group(1)
                    head, tail = os.path.split(fullPathAndName)
                    logging.debug(
                        'processing image original="{0}" head="{1}" tail="{2}" '
                        'logDirectory="{3}"'.format(
                            fullPathAndName, head, tail, logDirectory
                        )
                    )

                    # create the full path name
                    try:
                        if len(head) == 0:
                            # if logDirectory is empty then we try and find the file
                            if len(logDirectory) == 0:
                                fullPathAndName = self.findLatestInSessionDirectory(
                                    tail
                                )
                                logging.debug(
                                    "using latest file in suite {0}".format(
                                        fullPathAndName
                                    )
                                )

                            # otherwise we use the log directory
                            else:
                                logging.debug(
                                    'Using discovered log directory "{0}" on image '
                                    'file "{1}"'.format(logDirectory, tail)
                                )
                                fullPathAndName = os.path.join(logDirectory, tail)

                    except Exception as e:
                        logging.info("Error constructing full path {0}".format(str(e)))

                    # load the file (uuencoded) into the dictionary
                    try:
                        logging.info("Encoding image {0}".format(tail))
                        binaryData = open(fullPathAndName, "rb").read()
                        encodedData = base64.b64encode(binaryData).decode("utf-8")
                        uniquePath = self.constructUniquePathName(fullPathAndName)

                        logging.debug(
                            "image info to server: name={0}, fullpath={1}, "
                            "uniquePath={2}".format(tail, fullPathAndName, uniquePath)
                        )

                        images[tail] = {
                            "fullpath": fullPathAndName,
                            "uniquePath": uniquePath,
                            "contents": encodedData,
                        }

                    except Exception as e:
                        logging.info("Error encoding images {0}".format(str(e)))

            except TypeError:
                logging.info(
                    "returned value of {0} is not iterable".format(resultDict["Output"])
                )
            except Exception as e:
                logging.info("imageFileProcess() exception {0}".format(str(e)))

        return images

    def is_subdir(self, path, directory):
        try:
            path = os.path.realpath(path)
            directory = os.path.realpath(directory)
            relative = os.path.relpath(path, directory)
            return not relative.startswith(os.pardir + os.sep)
        except Exception as e:
            logging.info("Can't determine a relative path due to ({0})".format(str(e)))
            return False

    def constructUniquePathName(self, fullPathAndName):
        """
        In order to store a file on the server its name needs to be unique. We achieve
        this by creating a unique path based on the session name and the relative path
        of the Reports in ePF
        """
        # Remove everything before the suite name
        relPath = ""
        if self.is_subdir(self.sessionDirectory, fullPathAndName):
            relPath = os.path.relpath(fullPathAndName, self.sessionDirectory)
        else:
            # in this case make sure that the path is just the file name better than
            # nothing
            __head, relPath = os.path.split(fullPathAndName)

        uniquePath = os.path.join(self.uniqueSessionName, relPath)
        uniquePath = sub(r"\\", "/", uniquePath)
        logging.debug(
            "constructUniquePathName({0}):{1} uniqueSessionName={2}, "
            "relPath={3}".format(
                fullPathAndName, uniquePath, self.uniqueSessionName, relPath
            )
        )

        # replace last .png with _<unique>.png
        uniquePath = sub(r"(.*).png", r"\1_{0}.png".format(self.uniqId), uniquePath)
        self.uniqId = self.uniqId + 1

        return uniquePath

    def sessionLocationSetup(self, data):
        if ("cmd" in data) and ("arg" in data) and data["cmd"] == "StartSession":
            self.sessionDirectory = data["arg"]
            __head, tail = os.path.split(data["arg"])
            self.uniqueSessionName = (
                tail + "_" + datetime.now().strftime("%Y%m%d%H%M%S%f")
            )
            logging.debug(
                "sessionLocationSetup(data={0}):sessionDirectory={1}, "
                "uniqueSessionname={2}".format(
                    str(data), self.sessionDirectory, self.uniqueSessionName
                )
            )

    def replaceImageFileNames(self, images, result):
        full2short = {}
        short2unique = {}

        for m in finditer(r"\s(\S*\.png)\s", result):
            head, tail = os.path.split(m.group(1))

            # if there is a path before the file name then list it for long to short
            # replacement
            if len(head) > 0:
                full2short[m.group(1)] = tail

            # list all short name for short to unique replacemsnt
            if tail in images:
                short2unique[tail] = images[tail]["uniquePath"]

        # replace long name for short replacing spaces in names for %20
        for key, value in full2short.items():
            # substitute spaces with %20 so that they will be interpreted in the
            # html log
            result = sub(key, sub(r"\s", "%20", value), result)

        # add unique path to start replacing spaces in names for %20
        for key, value in short2unique.items():
            result = sub(key, sub(r"\s", "%20", value), result)
        return result

    def replaceImageFileNamesDict(self, images, resDict):
        full2short = {}
        short2unique = {}

        # loop through each element in the dictionary replacing images
        for k, v in resDict.items():
            if isinstance(v, dict):
                self.replaceImageFileNamesDict(images, v)

            elif isinstance(v, str):
                for m in finditer(r"\s(\S*\.png)\s", v):
                    head, tail = os.path.split(m.group(1))

                    # if there is a path before the file name then list it for long to
                    # short replacement
                    if len(head) > 0:
                        full2short[m.group(1)] = tail

                    # list all short name for short to unique replacemsnt
                    if tail in images:
                        short2unique[tail] = images[tail]["uniquePath"]

                # replace long name for short replacing spaces in names for %20
                for key, value in full2short.items():
                    # substitute spaces with %20 so that they will be interpreted in
                    # the html log
                    resDict[k] = sub(key, sub(r"\s", "%20", value), resDict[k])

                # add unique path to start replacing spaces in names for %20
                for key, value in short2unique.items():
                    resDict[k] = sub(key, sub(r"\s", "%20", value), resDict[k])
