import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from enum import IntEnum
from pathlib import Path

from eggplant.asset_manager.client import AssetDoesNotExistException, AssetType
from eggplant.asset_manager.client.ignore_rules import epf_suite_ignore_rules

from background_job import run_background_job
from folder import Folder
from severity import Severity
from suite import Suite
from eaiConstants import VAM_RUN_FOLDER


class SuiteState(IntEnum):
    LOCAL_ONLY = 0
    REMOTE_ONLY = 1
    LOCAL_OUT_OF_DATE = 2
    REMOTE_OUT_OF_DATE = 3
    UP_TO_DATE = 4


class SuiteManager(object):
    def __init__(self, asset_manager, root_folder, manage_suites_dialog):
        self.asset_manager = asset_manager
        self.root_folder = root_folder
        self.manage_suites_dialog = manage_suites_dialog
        self.remote_suites = {}
        self.local_suites = {}

    def start_background_job(self, callback, fn, *args, **kwargs):
        self.manage_suites_dialog.busy()

        def on_complete(result, exception):
            self.manage_suites_dialog.not_busy()
            callback(result, exception)

        run_background_job(self.manage_suites_dialog, on_complete, fn, *args, **kwargs)

    def load_suites(self):
        self.start_background_job(self.on_load_suites_complete, self.do_load_suites)

    def do_load_suites(self):
        self.remote_suites.clear()
        self.local_suites.clear()

        # Get all the EPF suites from the server
        assets = self.asset_manager.list_assets(
            asset_type_name=AssetType.EPF_SUITE.value
        )

        # For each suite, get the list of versions from the server
        # Do this in parallel for efficiency
        with ThreadPoolExecutor(max_workers=8) as executor:
            futures = {}

            # Submit each call to the thread pool
            for asset in assets:
                future = executor.submit(
                    self.asset_manager.list_asset_versions, asset.name
                )
                futures[future] = asset

            # Deal with each result as it completes
            for future in as_completed(futures):
                asset = futures[future]
                remote_versions = future.result()
                self.remote_suites[asset.name] = {
                    version.url: version for version in remote_versions
                }

        for suite_name in os.listdir(self.root_folder):
            if (
                os.path.isdir(os.path.join(self.root_folder, suite_name))
                and suite_name != VAM_RUN_FOLDER  # ignore the .run folder
            ):
                if suite_name in self.remote_suites:
                    local_version_url = self.asset_manager.get_current_version(
                        suite_name, self.root_folder
                    )

                    if local_version_url in self.remote_suites[suite_name]:
                        self.local_suites[suite_name] = self.remote_suites[suite_name][
                            local_version_url
                        ].name
                    else:
                        # Version of asset not in VAM
                        self.local_suites[suite_name] = "<UNKNOWN>"

                else:
                    # Asset not in VAM
                    self.local_suites[suite_name] = ""

        all_suite_names = set(self.remote_suites.keys()).union(self.local_suites.keys())

        suites = []

        for suite_name in sorted(all_suite_names, key=str.casefold):
            remote_version = self.get_latest_version(suite_name)
            local_version = self.local_suites.get(suite_name)

            suites.append(Suite(suite_name, local_version, remote_version))

        return suites

    def on_load_suites_complete(self, suites, exception):
        if exception is not None:
            self.manage_suites_dialog.set_suite_information(
                "Unable to load suites: {}".format(exception), severity=Severity.ERROR
            )
        else:
            self.manage_suites_dialog.populate_suites(suites)

    def scan_suite(self, suite_name):
        self.start_background_job(
            self.on_scan_complete, self.do_scan, suite_name,
        )

    def do_scan(self, suite_name):

        # Create a top level Folder object and then populate it with
        # all the subfolders/files
        top_level_folder = Folder("")

        if suite_name not in self.local_suites:
            state = SuiteState.REMOTE_ONLY
        else:
            try:
                changed_files = self.asset_manager.report_changes(
                    suite_name, self.root_folder
                )

                for key, file_list in changed_files.items():
                    for path in file_list:
                        self.add_file_to_folder(path, top_level_folder, key)

                latest_version = self.get_latest_version(suite_name)

                if latest_version.name != self.local_suites[suite_name]:
                    state = SuiteState.LOCAL_OUT_OF_DATE
                elif not top_level_folder.is_empty():
                    state = SuiteState.REMOTE_OUT_OF_DATE
                else:
                    state = SuiteState.UP_TO_DATE

            except AssetDoesNotExistException:
                # Asset not in the VAM
                state = SuiteState.LOCAL_ONLY

                # Find all the files that would be added,
                # using the standard EPF suite ignore rules
                suite_path = Path(self.root_folder, suite_name)
                for path in epf_suite_ignore_rules.get_noticed_files(str(suite_path)):
                    relative_path = path.relative_to(suite_path)
                    self.add_file_to_folder(
                        str(relative_path), top_level_folder, "added"
                    )

        return suite_name, state, top_level_folder

    def add_file_to_folder(self, path, top_level_folder, key):
        # split the file path into it's components
        components = os.path.normpath(path).split(os.sep)

        # Create any folders in the path
        folder = top_level_folder
        for component in components[:-1]:
            folder = folder.add_folder(component)

        # Add the file to the right folder
        folder.add_file(components[-1], key)

    def on_scan_complete(self, result, exception):
        if exception is not None:
            self.manage_suites_dialog.set_suite_information(
                "Unable to get file changes: {}".format(exception),
                severity=Severity.ERROR,
            )
        else:
            suite_name, state, folder = result
            self.manage_suites_dialog.populate_deltas(suite_name, state, folder)

    def upload_suite(self, suite_name, new_version_name):
        self.start_background_job(
            self.on_uploaded_suite, self.do_upload_suite, suite_name, new_version_name,
        )

    def do_upload_suite(self, suite_name, new_version_name):
        if suite_name in self.remote_suites:
            # Update existing asset
            new_version = self.asset_manager.save_and_publish_new_asset_version(
                suite_name, self.root_folder, new_version_name
            )
        else:
            # Create new asset
            new_version = self.asset_manager.create_asset(
                os.path.join(self.root_folder, suite_name),
                version_name=new_version_name,
            )

        return new_version

    def on_uploaded_suite(self, new_version, exception):
        if exception is not None:
            self.manage_suites_dialog.set_suite_information(
                "Failed to upload suite: {}".format(exception), severity=Severity.ERROR
            )
        else:
            self.manage_suites_dialog.set_suite_information(
                "Suite successfully uploaded", severity=Severity.INFO
            )
            self.load_suites()

    def download_suite(self, suite_name, should_overwrite_local_changes):
        self.start_background_job(
            self.on_downloaded_suite,
            self.do_download_suite,
            suite_name,
            should_overwrite_local_changes,
        )

    def do_download_suite(self, suite_name, should_overwrite_local_changes):
        # TODO use the AssetController class?
        version = self.asset_manager.download_asset(
            assets_dir=self.root_folder,
            asset_name=suite_name,
            version_name=None,  # will download latest version
            overwrite=should_overwrite_local_changes,
        )

        return suite_name, version.name

    def on_downloaded_suite(self, result, exception):
        if exception is not None:
            self.manage_suites_dialog.set_suite_information(
                "Failed to download suite: {}".format(exception),
                severity=Severity.ERROR,
            )
        else:
            suite_name, version_name = result
            self.manage_suites_dialog.set_suite_information(
                "Downloaded version {} of suite {}".format(version_name, suite_name),
                severity=Severity.INFO,
            )
            self.load_suites()

    def get_existing_version_names(self, suite_name):
        if suite_name not in self.remote_suites:
            return []

        return [version.name for version in self.remote_suites[suite_name].values()]

    def get_latest_version(self, suite_name):
        versions = self.remote_suites.get(suite_name, {})
        if len(versions) == 0:
            return None

        return sorted(versions.values(), key=lambda v: v.created)[-1]

    def suggest_version_name(self, suite_name):
        if suite_name not in self.remote_suites:
            return "version-1"

        return "version-{}".format(len(self.remote_suites[suite_name]) + 1)

    def get_local_suite_names(self):
        return self.local_suites.keys()
