import logging

import jwt
import requests
from eggplant.utils.auth_token import AuthToken


class TokenValidationException(Exception):
    pass


def verify_token(server, is_ssl, token):
    """
    Make a call to the server to verify that the token is valid,
    """
    address = "{}://{}".format("https" if is_ssl else "http", server)

    try:
        logging.info("Verifying token against server %s", address)

        # Make a call to AI API to verify that the token is valid
        response = requests.get(
            "{}/ai/auth".format(address),
            headers={"Authorization": "Bearer {}".format(token)},
        )
        response.raise_for_status()
        logging.info("Token verified")
    except requests.ConnectionError:
        raise TokenValidationException(
            "Unable to connect to server at {}".format(address)
        )
    except requests.HTTPError as e:
        if e.response.status_code == 401:
            logging.error("Invalid token")
            raise TokenValidationException("Invalid token")
        if e.response.status_code == 404:
            # AI API is probably starting up
            logging.error("Auth endpoint not found")
            raise TokenValidationException("Auth endpoint not found")

        raise


def verify_permissions(token, is_design_mode):
    if not is_design_mode and not is_user_an_admin(token):
        raise TokenValidationException(
            "You need to be an administrator to use the agent in RUN mode"
        )


def is_user_an_admin(token):
    # Deserialize the token
    decoded_token = jwt.decode(token, verify=False)
    auth_token = AuthToken.load(decoded_token)

    return auth_token.user.role == "Admin"
