import logging
import socket
import glob
import os
import io
import base64
import time
import uuid

from json import loads, dumps

from eggDriveInterface import EggDriveStarter
from webSocketConnection import WebSocketConnection
from imageHandling import ImageHandling
from utilities import Utilities
from base64 import b64encode, b64decode
from PIL import Image
from asset_controller import DownloadResult, DownloadStatus

UUID = str(uuid.uuid4())
MAX_BACKOFF_TIME = 32


class Top(ImageHandling, Utilities):
    """Top level of the TT agent
    """

    def __init__(self, args, asset_controller, on_stop=None):
        """
        Constructor
        args is a Namespace object returned from the argparse module:
        https://docs.python.org/3/library/argparse.html#argparse.Namespace
        """

        ImageHandling.__init__(self)
        # Need to create a structure including the eggDriveInformation
        self.ePFIdent = {}
        self.ePFIdent["host"] = args.server  # The TT server host name
        # The TCP/IP port used for eggDrive
        self.ePFIdent["port"] = args.port
        self.ePFIdent["executionMode"] = "DESIGN" if args.designMode else "RUN"
        self.driveExec = args.driveExec
        self.epfgui = args.epfgui
        self.ePFIdent["suiteRootFolder"] = args.suiteRootFolder
        self.shutting_down = False
        self.connection_backoff = 1
        self.ePFIdent["uuid"] = UUID

        # The protocol is conditional
        self.ePFIdent["protocol"] = "wss://" if args.ssl is True else "ws://"

        # if the name parameter was set then we use that instead of myIPAddr and
        # myFQDName
        if len(args.name) > 0:
            self.ePFIdent["agentName"] = args.name
            logging.info(
                "Connecting {0} to server {1}{2} with eggDrive running on "
                "port {3}".format(
                    self.ePFIdent["agentName"],
                    self.ePFIdent["protocol"],
                    self.ePFIdent["host"],
                    self.ePFIdent["port"],
                )
            )
        else:
            try:
                self.ePFIdent["myIPAddr"] = [
                    l
                    for l in (
                        [
                            ip
                            for ip in socket.gethostbyname_ex(socket.gethostname())[2]
                            if not ip.startswith("127.")
                        ][:1],
                        [
                            [
                                (
                                    s.connect(("8.8.8.8", 53)),
                                    s.getsockname()[0],
                                    s.close(),
                                )
                                for s in [
                                    socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                                ]
                            ][0][1]
                        ],
                    )
                    if l
                ][0][0]
            except socket.gaierror:
                self.ePFIdent["myIPAddr"] = "127.0.0.1"
            # get the fully qualified domain name
            self.ePFIdent["myFQDName"] = socket.getfqdn()
            logging.info(
                "Connecting {0} ({1}) to server {2}{3} with eggDrive "
                "running on port {4}".format(
                    self.ePFIdent["myIPAddr"],
                    self.ePFIdent["myFQDName"],
                    self.ePFIdent["protocol"],
                    self.ePFIdent["host"],
                    self.ePFIdent["port"],
                )
            )

        self.createEggDriveStarter(args)

        # Make an instance of the webSocket interface
        self.prepare_socket()

        self.asset_controller = asset_controller
        self.on_stop = on_stop

    def run(self):
        """Start any process that need to be started
        """
        self.shutting_down = False
        self.asset_controller.start()
        self.webSocketConnection.run()

    def createEggDriveStarter(self, args):
        # Make an instance of the eggDrive interface
        self.eggDriveInterface = EggDriveStarter(
            self.driveExec,
            self.ePFIdent["port"],
            self.epfgui,
            args.cleanepf,
            args.debug,
            args.startepf,
        )

    def prepare_socket(self):
        self.webSocketConnection = WebSocketConnection(
            ePFIdent=self.ePFIdent,
            messageCallback=self.wsMessageReceived,
            openCallback=self.wsOpen,
            closureCallback=self.wsClosed,
        )

    def cleanDictForJSON(self, dictnry):
        for key, value in dictnry.items():
            if isinstance(value, dict):
                self.cleanDictForJSON(value)
            elif key == "RunDate":  # change RunDate to a string
                dictnry[key] = str(value)

    def wsOpen(self):
        self.connection_backoff = 1

    def wsClosed(self):
        if self.shutting_down:
            return

        self.shutting_down = True
        del self.webSocketConnection

        logging.debug("Retrying in %d second(s)", self.connection_backoff)
        time.sleep(self.connection_backoff)

        if self.connection_backoff < MAX_BACKOFF_TIME:
            self.connection_backoff *= 2

        self.prepare_socket()
        self.run()

    def wsMessageReceived(self, msg):
        """called when we receive a websocket message.
        This is the main part of Top and handles passing commands and responses back and
        forth between the ePF interface and the websocket interface
        """

        try:
            data = loads(msg)
            logging.debug("Received from server\n{}".format(dumps(data, indent=2)))
        except ValueError:  # as e:
            logging.debug('Received from server "{}"'.format(msg))
            return

        # send the deferred id and results in a dictionary
        ident = 0
        if "eggDriveMsgIdent" in data:
            ident = data["eggDriveMsgIdent"]

        # Deal with errors sent from the server (we just quit)
        if "error" in data:
            logging.error("ERROR {0}".format(data["error"]))
            self.close()
            return

        # run any script
        elif ("cmd" in data) and (data["cmd"] == "script"):

            result, resDict = self.eggDriveInterface.send("script", data["arg"])

            # following 2 lines added to return json format
            self.cleanDictForJSON(resDict)
            result = dumps(resDict)
            result = b64encode(result.encode(), b"-_")

            if "deferred" in data:
                self.webSocketConnection.send(
                    {
                        "deferred": data["deferred"],
                        "eggDriveMsgIdent": ident,
                        "result": result.decode(),
                    }
                )
            else:
                self.webSocketConnection.send(
                    {"result": result.decode(), "eggDriveMsgIdent": ident}
                )
            return

        elif (
            ("cmd" in data) and (data["cmd"] == "scriptFile") and "path" in data["arg"]
        ):
            scriptFile = os.path.join(data["arg"]["path"], data["arg"]["scriptFile"])
            script = ""
            with open(scriptFile, "r") as f:
                script = str(f.read())
            result, resDict = self.eggDriveInterface.send(
                "script", {"path": data["arg"]["path"], "script": script}
            )

            if "deferred" in data:
                self.webSocketConnection.send(
                    {
                        "deferred": data["deferred"],
                        "eggDriveMsgIdent": ident,
                        "result": result,
                    }
                )
            else:
                self.webSocketConnection.send(
                    {"result": result, "eggDriveMsgIdent": ident}
                )
            return

        # general purpose commands
        elif ("cmd" in data) and ("arg" in data):
            # if this is an execute then we need to decode the base 64 args
            logging.info("Performing CMD {0}".format(data["cmd"]))

            if data["cmd"] == "ExecuteWithTimeout":
                data["arg"] = b64decode(data["arg"], b"-_")

            self.sessionLocationSetup(data)
            result, resDict = self.eggDriveInterface.send(data["cmd"], data["arg"])

            if data["cmd"] == "GetScreen":
                imageInBytes = sutScreenSave(resDict)
                imageData = base64.b64encode(imageInBytes).decode("utf-8")

                if "deferred" in data:
                    self.webSocketConnection.send(
                        {
                            "deferred": data["deferred"],
                            "eggDriveMsgIdent": ident,
                            "sutscreen": imageData,
                        }
                    )
                else:
                    self.webSocketConnection.send(
                        {"sutscreen": imageData, "eggDriveMsgIdent": ident}
                    )
                return

            if data["cmd"] == "CaptureImage":
                if "deferred" in data:
                    self.webSocketConnection.send(
                        {
                            "deferred": data["deferred"],
                            "eggDriveMsgIdent": ident,
                            "result": resDict,
                        }
                    )
                else:
                    self.webSocketConnection.send(
                        {"result": resDict, "eggDriveMsgIdent": ident}
                    )
                return

            images = self.imageFileProcess(resDict)

            # replace image names in the file with the server based ones
            try:
                result = self.replaceImageFileNames(images, result)
                self.replaceImageFileNamesDict(images, resDict)
            except Exception:
                logging.exception("Failed to replace images")

            # send the result back via the websocket
            if len(result) > 0:
                # following 2 lines added to return json format
                self.cleanDictForJSON(resDict)
                result = dumps(resDict)

                result = b64encode(result.encode(), b"-_")
                # result = dumps(resDict, ensure_ascii=True)
                self.webSocketConnection.send(
                    {
                        "msg": result.decode(),
                        "eggDriveMsgIdent": ident,
                        "images": images,
                    }
                )
            else:
                if "deferred" in data:
                    self.webSocketConnection.send(
                        {
                            "deferred": data["deferred"],
                            "eggDriveMsgIdent": ident,
                            "msg": result,
                        }
                    )
                else:
                    self.webSocketConnection.send(
                        {"msg": "No data here", "eggDriveMsgIdent": ident}
                    )

        # Perform local actions (e.g. listing sensetal/python/csv files)
        elif "action" in data:
            logging.info("Performing Action {0}".format(data["action"]))
            results = []

            # return a list of sensetalk files
            if data["action"] == "listSensetalk":
                files = glob.iglob(data["path"] + "/**/*.script", recursive=True)
                for f in files:
                    # _, fileName = self.fileNameFromPath(f)
                    fileName = self.toForwardSlash(os.path.relpath(f, data["path"]))
                    fileResult = {}  # A dictionary for each file
                    fileResult["scriptFile"] = fileName
                    # b64encode(contents.encode(), b'-_').decode()
                    fileResult["contents"] = open(f, "r", encoding="utf-8-sig").read()
                    fileResult["hash"] = self.calculateCRC(fileResult["contents"])
                    results.append(fileResult)

                files = []
                for r in results:
                    files.append(r["scriptFile"])

            # return a list of csv files
            elif data["action"] == "listCsv":
                for p in data["path"]:
                    files = glob.glob(p + "/*.csv")
                    for f in files:
                        _, fileName = self.fileNameFromPath(f)
                        fileResult = {}  # A dictionary for each file
                        fileResult["csvFile"] = fileName
                        # b64encode(open(f, 'r').read().encode(), b'-_').decode()
                        fileResult["contents"] = open(
                            f, "r", encoding="utf-8-sig"
                        ).read()
                        fileResult["hash"] = self.calculateCRC(fileResult["contents"])
                        results.append(fileResult)

            # return a list of csv files
            elif data["action"] == "fetchFile":
                paths = data["path"].split("|")
                paths[-1], fileName = self.fileNameFromPath(paths[-1])
                for p in paths:
                    files = glob.glob(p + "/" + fileName)

                    for f in files:
                        _, fileName = self.fileNameFromPath(f)
                        fileResult = {}  # A dictionary for each file
                        fileResult["file"] = fileName
                        # b64encode(open(f, 'r').read().encode(), b'-_').decode()
                        fileResult["contents"] = open(
                            f, "r", encoding="utf-8-sig"
                        ).read()
                        fileResult["hash"] = self.calculateCRC(fileResult["contents"])

                        results.append(fileResult)

                    # Pass back the file contents raw as this will be going to
                    # Eggplant AI command line
                    # TODO this looks wrong why overwrite results here
                    results = results[0]["contents"]

            elif data["action"] == "readFile":
                try:
                    with open(data["arg"]["path"], "r", encoding="utf-8-sig") as file:
                        results = {"result": "success", "contents": file.read()}
                except Exception:
                    results = {"result": "error"}

            elif data["action"] == "saveFile":
                # Create the directory if it doesn't exist
                os.makedirs(os.path.dirname(data["arg"]["path"]), exist_ok=True)
                try:
                    with open(data["arg"]["path"], "w") as file:
                        file.write(data["arg"]["content"])
                    results = {"result": "success"}
                except Exception:
                    results = {"result": "error"}

            elif data["action"] == "delFile":
                try:
                    os.remove(data["arg"]["path"])
                    results = {"result": "success"}
                except Exception:
                    results = {"result": "error"}

            elif data["action"] == "listFiles":
                try:
                    files = glob.iglob(data["arg"]["path"] + "/**/*", recursive=True)
                    results = []
                    for f in files:
                        results.append(
                            {
                                "path": str(f),
                                "relpath": str(f)
                                .replace(data["arg"]["path"], "")
                                .replace("\\", "/")
                                .strip("/\\"),
                                "type": "file" if os.path.isfile(f) else "dir",
                            }
                        )
                except Exception:
                    results = {"result": "error"}

            elif data["action"] == "download_asset":
                self.asset_controller.download_asset(
                    data["arg"]["path"], data["arg"]["version_id"]
                )
                results = DownloadResult(DownloadStatus.IN_PROGRESS).to_dict()
            elif data["action"] == "download_asset_progress":
                results = self.asset_controller.get_result()
            elif data["action"] == "cancel_download_asset":
                self.asset_controller.cancel_download()
                results = DownloadResult(DownloadStatus.CANCELLED).to_dict()
            if "deferred" in data:
                self.webSocketConnection.send(
                    {
                        "deferred": data["deferred"],
                        "eggDriveMsgIdent": ident,
                        "results": results,
                    }
                )
            elif len(results) > 0:
                self.webSocketConnection.send(
                    {"msg": results, "eggDriveMsgIdent": ident}
                )

    def close(self):
        """ Close anything that needs closing
        """
        logging.info("Agent close requested")
        self.shutting_down = True
        self.webSocketConnection.close(self.webSocketConnection.ws)
        self.eggDriveInterface.close()
        self.asset_controller.stop()

        if self.on_stop is not None:
            self.on_stop()


def sutScreenSave(resDict):
    """
    From the resDict data, save the image in png format and return the
    generated filename
    Returns the byte representation of the image in png format
    """
    value = None
    with Image.frombytes(
        "RGB",
        (resDict["Result"]["width"], resDict["Result"]["height"]),
        resDict["Result"]["data"],
    ) as im:
        with io.BytesIO() as byteIO:
            im.save(byteIO, format="PNG")
            value = byteIO.getvalue()
    return value
