"""
Created on 17 Jan 2017

@author: giles.hall
"""
import json
import logging
import ssl

import certifi
import websocket


class WebSocketConnection(object):
    """An instance of the websocket connection for communication back to the model server
    """

    def __init__(
        self, ePFIdent, messageCallback, openCallback, closureCallback
    ):  # e.g. server = 127.0.0.1
        self.serverName = ePFIdent["protocol"] + ePFIdent["host"] + "/ttws"
        self.ePFIdent = ePFIdent
        self.messageCallback = messageCallback
        self.closureCallback = closureCallback
        self.openCallback = openCallback

        context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        context.load_default_certs()

        logging.info("Loading SSL certificates from file: {}".format(certifi.where()))
        logging.info("Loaded {} SSL certificates".format(len(context.get_ca_certs())))

        # websocket.enableTrace(True)
        self.ws = websocket.WebSocketApp(
            self.serverName,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
        )
        self.ws.on_open = self.on_open

    def on_message(self, ws, message):
        """Called every time a message is received on the websocket
        """
        try:
            self.messageCallback(message)
        except Exception as e:
            logging.error(
                (
                    "Unhandled exception {} while running on_message handler for "
                    "message {}"
                ).format(type(e).__name__, message)
            )
            logging.exception(e)

    def on_error(self, ws, error):
        """Called if we get an error from the websocket
        """
        logging.info(error)

    def close(self, ws):
        """ close the websocket when requested
        """
        logging.info("WebSocketConnection.close() requested")
        self.ws.close()

    def on_close(self, ws):
        """Called when the websocket is closed
        """
        logging.info("Websocket connection closed")
        self.closureCallback()

    def on_open(self, ws):
        """Called when the websocket is opened
        """
        logging.info("websocket opened to server " + self.serverName)
        # send a message to the server giving our details
        logging.info("Sending ident to server")
        # just registration flag in addition to the default parameters
        self.send({"register": True})
        self.openCallback()

    def send(self, message_dictionary):
        """Send data as JSON via the websocket
        """
        # Add agent identification info to the message
        message_dictionary.update(self.ePFIdent)

        # Convert message to string
        msg = json.dumps(
            message_dictionary, separators=(",", ":"), ensure_ascii=True, indent=2
        )

        logging.debug("Sending to server: %s", msg)
        self.ws.send(msg)

    def run(self):
        """ run the websocket interface
        """
        self.ws.run_forever(ping_interval=30, ping_timeout=20)
